/*
 * Copyright 2024 KylinSoft Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include "bgem3embedder.h"

#include <algorithm>
#include <iostream>

static const char* MODEL_FOLDER = "/home/wangweinan/models/bge-m3-onnx";
static const char* MODEL_PATH = "/home/wangweinan/models/bge-m3-onnx/model.onnx";

Bgem3Embedder::Bgem3Embedder()
    : tokenizer_(MODEL_FOLDER)
    , session_(Ort::Env(ORT_LOGGING_LEVEL_WARNING, "Default"),
               MODEL_PATH, Ort::SessionOptions(nullptr))
{}

int Bgem3Embedder::longestTokensSize(const std::vector<Token> &tokens) const
{
    int longestLength = 0;

    for (const auto& mapItem : tokens) {
        for (const auto& pair : mapItem) {
            longestLength = std::max<int>(longestLength, pair.second.size());
        }
    }

    return longestLength;
}

std::vector<Ort::Value> Bgem3Embedder::createInputTensors(Token &inputTokens, int batch, const std::vector<const char*>& inputNames) const
{
    Ort::MemoryInfo memoryInfo = Ort::MemoryInfo::CreateCpu(OrtArenaAllocator, OrtMemTypeCPU);
    std::vector<Ort::Value> inputTensors;

    for (int i = 0; i < inputNames.size(); ++i) {
        // tokens 的顺序需要和 names 一一对应
        const std::string& inputName = inputNames[i];
        std::vector<int64_t>& inputValue = inputTokens[inputName];
        const int length = inputValue.size() / batch;
        const std::vector<int64_t> shape = { batch, length };

        Ort::Value inputTensor = Ort::Value::CreateTensor<int64_t>(
            memoryInfo, inputValue.data(), inputValue.size(), shape.data(), shape.size());
        inputTensors.emplace_back(std::move(inputTensor));
    }

    return inputTensors;
}

Ort::Value Bgem3Embedder::createOutputTensor(std::vector<float>& outputs) const
{
    const int batch = outputs.size() / 1024;
    const std::vector<int64_t> outputShape = { batch, 1024 };
    Ort::MemoryInfo memoryInfo = Ort::MemoryInfo::CreateCpu(OrtArenaAllocator, OrtMemTypeCPU);

    Ort::Value outputTensor = Ort::Value::CreateTensor<float_t>(
        memoryInfo, outputs.data(), outputs.size(), outputShape.data(), outputShape.size());

    return outputTensor;
}

void Bgem3Embedder::resizeTokens(std::vector<Token> &tokens, size_t maxLength) const
{
    for (auto& token : tokens) {
        for (auto& pair : token) {
            std::vector<int64_t>& values = pair.second;
            if (values.size() < maxLength) {
                values.resize(maxLength, 0);
            }
        }
    }
}

Token Bgem3Embedder::mergeTokens(const std::vector<Token> &tokens) const
{
    Token result;

    for (const auto& token : tokens) {
        for (const auto& pair : token) {
            const std::string& key = pair.first;
            const std::vector<int64_t>& value = pair.second;

            std::vector<int64_t>& resultValue = result[key];
            resultValue.reserve(resultValue.size() + value.size());
            resultValue.insert(resultValue.end(), value.begin(), value.end());
        }
    }

    return result;
}

Token Bgem3Embedder::tokenize(const std::vector<std::string> &texts)
{
    std::vector<Token> tokenList;
    for (const std::string& text : texts) {
        tokenList.emplace_back(tokenizer_.tokenize(text));
    }

    // 在进行批处理时首先要将 tokens 对齐，让所有 tokens 长度一致，
    // 对于不够长的可以补 0 并在 attention_mask 进行标记
    // 然后需要将所有token合并成一个 token, 如果只合并了 token 没有对齐会导致推理结果错误
    const int maxLength = longestTokensSize(tokenList);
    resizeTokens(tokenList, maxLength);
    return mergeTokens(tokenList);
}

std::vector<std::vector<float>> Bgem3Embedder::encodeTexts(const std::vector<std::string>& texts)
{
    if (texts.empty()) {
        return std::vector<std::vector<float>>();
    }

    Token tokens = tokenize(texts);

    const int batch = static_cast<int>(texts.size());
    const std::vector<const char*> inputNames = {"input_ids", "attention_mask"};
    const std::vector<const char*> outputNames = {"dense_vecs"};

    std::vector<float> output(batch * 1024);

    std::vector<Ort::Value> inputTensors = createInputTensors(tokens, batch, inputNames);

    Ort::Value outputTensor = createOutputTensor(output);

    Ort::RunOptions runOpts(nullptr);
    session_.Run(runOpts, inputNames.data(), inputTensors.data(), inputTensors.size(),
                 outputNames.data(), &outputTensor, 1);

    std::vector<std::vector<float>> outputs;
    for (int i = 0; i < batch; ++i) {
        std::vector<float> temp(output.begin() + i * 1024, output.begin() + (i + 1) * 1024);
        outputs.push_back(temp);
    }

    return outputs;
}
