/*
 * Copyright 2024 KylinSoft Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include "services/speechprocessor.h"

#include <iostream>
#include <tuple>
#include <kylin-ai/ai-engine/large-model/cloudspeechengine.h>
#include <openssl/bio.h>
#include <openssl/evp.h>
#include <iomanip>
#include "config/engineconfiguration.h"
#include "services/threadpool/async.h"

const std::string SpeechProcessor::objectPath_ =
    "/org/openkylin/aisdk/speechprocessor";

SpeechProcessor::SpeechProcessor(
        GDBusConnection &connection,
        ai_engine::AiEnginePluginManager &aiEnginePluginManager)
    : connection_(connection), aiEnginePluginManager_(aiEnginePluginManager) {
    exportSkeleton();
}

SpeechProcessor::~SpeechProcessor() { unexportSkeleton(); }

void SpeechProcessor::stopProcess() {
    if (speechEngine_) {
        speechEngine_->stopContinuousSynthesis();
        speechEngine_->stopContinuousRecognition();
    }

    std::lock_guard<std::mutex> locker(mutex_);
    stopped = true;
}

void SpeechProcessor::exportSkeleton() {
    delegate_ = aisdk_speech_processor_skeleton_new();
    connectSignal();

    GError *error = nullptr;

    isExported_ = g_dbus_interface_skeleton_export(
        G_DBUS_INTERFACE_SKELETON(delegate_), &connection_, objectPath_.c_str(),
        &error);
    if (!isExported_) {
        g_printerr("Error creating server at address %s: %s\n",
                   objectPath_.c_str(), error->message);
        g_error_free(error);
    }
}

std::string SpeechProcessor::base64decode(const std::string &input) {
    BIO *bio, *b64;
    char *buffer = new char[input.size()];
    memset(buffer, 0, input.size());

    b64 = BIO_new(BIO_f_base64());
    BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);
    bio = BIO_new_mem_buf(input.c_str(), input.size());
    bio = BIO_push(b64, bio);

    BIO_read(bio, buffer, input.size());

    BIO_free_all(bio);

    std::string result(buffer, buffer + input.size());

    delete[] buffer;
    return result;
}

void SpeechProcessor::emitRecognizingResult(
        std::shared_ptr<SpeechProcessor> speechProcessor,
        ai_engine::lm::speech::RecognitionResult result,
        const gchar *signalName) {
    GVariantBuilder builder;
    g_variant_builder_init (&builder, G_VARIANT_TYPE_TUPLE);
    g_variant_builder_add(&builder, "s", result.text.c_str());
    g_variant_builder_add(&builder, "i", result.text.length());
    g_variant_builder_add(&builder, "i", result.errorCode);

    g_dbus_connection_emit_signal(&(speechProcessor->connection_),
                                  "org.openkylin.aisdk.speechprocessor",
                                  "/org/openkylin/aisdk/speechprocessor",
                                  "org.openkylin.aisdk.speechprocessor",
                                  signalName,
                                  g_variant_builder_end(&builder),
                                  nullptr);
}

void SpeechProcessor::emitSynthesizingResult(
        std::shared_ptr<SpeechProcessor> speechProcessor,
        ai_engine::lm::speech::SynthesisResult result,
        const gchar * signalName) {
    // 创建包含字节数组的 GVariant
    GVariant *byte_array_variant = g_variant_new_from_data(
        G_VARIANT_TYPE_BYTESTRING,
        result.audioData.data(),
        result.audioData.size(),
        TRUE, NULL, NULL);

    GVariantBuilder builder;
    g_variant_builder_init (&builder, G_VARIANT_TYPE_TUPLE);
    g_variant_builder_open (&builder, G_VARIANT_TYPE_ARRAY);
    g_variant_builder_add_value(&builder, byte_array_variant);
    g_variant_builder_close (&builder);

    g_variant_builder_add(&builder, "s", result.audioFormat.c_str());
    g_variant_builder_add(&builder, "i", result.audioRate);
    g_variant_builder_add(&builder, "i", result.audioChannel);
    g_variant_builder_add(&builder, "i", result.errorCode);

    g_dbus_connection_emit_signal(&(speechProcessor->connection_),
                                  "org.openkylin.aisdk.speechprocessor",
                                  "/org/openkylin/aisdk/speechprocessor",
                                  "org.openkylin.aisdk.speechprocessor",
                                  signalName,
                                  g_variant_builder_end(&builder),
                                  nullptr);
}

void SpeechProcessor::initEngine(const std::string &engineName, const std::string &appId,
                                 const std::string &apiKey, const std::string &secretKey) {
    speechEngine_ = aiEnginePluginManager_.createSpeechEngine(engineName);
    if (speechEngine_->isCloud()) {
        static_cast<ai_engine::lm::speech::AbstractCloudSpeechEngine *>(speechEngine_.get())->setAppId(appId);
        static_cast<ai_engine::lm::speech::AbstractCloudSpeechEngine *>(speechEngine_.get())->setApiKey(apiKey);
        static_cast<ai_engine::lm::speech::AbstractCloudSpeechEngine *>(speechEngine_.get())->setSecretKey(secretKey);
    }
}

void onHandleInitEngine(AisdkSpeechProcessor *delegate,
                        GDBusMethodInvocation *invocation, gchar *engineName, gchar *appId,
                        gchar *apiKey, gchar *secretKey, gpointer userData) {
    auto *speechProcessor = static_cast<SpeechProcessor *>(userData);
    if (!speechProcessor) {
        return;
    }
    speechProcessor->promptTemplate_.reset(
        new config::PromptTemplate(config::promptTemplateConfigFile(engineName),
                                   config::promptFilePath(engineName))
    );
    speechProcessor->initEngine(engineName, appId, apiKey, secretKey);
    speechProcessor->speechEngine_->initContinuousRecognitionModule();
    speechProcessor->speechEngine_->initSynthesizeOnceModule();
}

bool onHandleStartRecognizingAudio(AisdkSpeechProcessor *delegate,
                                   GDBusMethodInvocation *invocation,
                                   gchar *audioProperties,
                                   gpointer userData) {
    auto *speechProcessor = static_cast<SpeechProcessor *>(userData);

    speechProcessor->speechEngine_->startContinuousRecognition(audioProperties);

    aisdk_speech_processor_complete_start_recognizing_audio(
        delegate, invocation);

    return true;
}

bool onHandleInputAudioData(AisdkSpeechProcessor *delegate,
                            GDBusMethodInvocation *invocation,
                            gchar *audioData, gint audioDataLength,
                            gpointer userData) {
    std::shared_ptr<SpeechProcessor> speechProcessor =
            static_cast<SpeechProcessor *>(userData)->getSharedObject();

    std::string data { audioData, (unsigned long)audioDataLength };
    std::string decodeData = speechProcessor->base64decode(data);
    std::vector<char> inputData { decodeData.data(), decodeData.data() + decodeData.size() };

    cpr::async([speechProcessor, inputData]() {
        speechProcessor->speechEngine_->setRecognizingCallback(
                    [speechProcessor](ai_engine::lm::speech::RecognitionResult result){
            std::cout << "service---onHandleInputAudioData--callback" << std::endl;
            {
                std::lock_guard<std::mutex> locker(speechProcessor->mutex_);
                if (speechProcessor->stopped) {
                    return;
                }
            }
            speechProcessor->emitRecognizingResult(
                speechProcessor, result, "RecognitionAudioResult");
        });
        speechProcessor->speechEngine_->setRecognizedCallback(
                    [](ai_engine::lm::speech::RecognitionResult result){
        });
        speechProcessor->speechEngine_->setSynthesizedCallback(
                    [](ai_engine::lm::speech::ResultType result) {
        });

        speechProcessor->speechEngine_->writeContinuousRecognitionAudioData(inputData);
    });

    aisdk_speech_processor_complete_input_audio_data(
        delegate, invocation, "input_audio_data", strlen("input_audio_data"), 0);
    return true;
}

bool onHandleStopRecognizingAudio(AisdkSpeechProcessor *delegate,
                                  GDBusMethodInvocation *invocation,
                                  gpointer userData) {
    std::shared_ptr<SpeechProcessor> speechProcessor =
            static_cast<SpeechProcessor *>(userData)->getSharedObject();

    cpr::async([speechProcessor]() {
        speechProcessor->speechEngine_->setRecognizedCallback(
                    [speechProcessor](ai_engine::lm::speech::RecognitionResult result){
            std::cout << "service---onHandleStopRecognizingAudio--callback" << std::endl;
            {
                std::lock_guard<std::mutex> locker(speechProcessor->mutex_);
                if (speechProcessor->stopped) {
                    return;
                }
            }
            speechProcessor->emitRecognizingResult(
                speechProcessor, result, "StopRecognitionAudioResult");
        });

        speechProcessor->speechEngine_->stopContinuousRecognition();
    });

    aisdk_speech_processor_complete_stop_recognizing_audio(
        delegate, invocation);
    return true;
}

bool onHandleRecognizingCompleteAudio(AisdkSpeechProcessor *delegate,
                                      GDBusMethodInvocation *invocation,
                                      gchar *audioProperties,
                                      gchar *audioData,
                                      gint audioDataLength,
                                      gpointer userData) {
    std::shared_ptr<SpeechProcessor> speechProcessor =
            static_cast<SpeechProcessor *>(userData)->getSharedObject();

    std::string data { audioData, (unsigned long)audioDataLength };
    std::string decodeData = speechProcessor->base64decode(data);
    std::vector<char> inputData { decodeData.data(), decodeData.data() + decodeData.size() };
    std::string params { audioProperties };

    cpr::async([speechProcessor, inputData, params]() {
        speechProcessor->speechEngine_->setRecognizingCallback(
                    [speechProcessor](ai_engine::lm::speech::RecognitionResult result){
            std::cout << "service---onHandleRecognizingCompleteAudio--callback" << std::endl;
            {
                std::lock_guard<std::mutex> locker(speechProcessor->mutex_);
                if (speechProcessor->stopped) {
                    return;
                }
            }
            speechProcessor->emitRecognizingResult(
                        speechProcessor, result, "RecognitionAudioOnceResult");
        });
        speechProcessor->speechEngine_->setRecognizedCallback(
                    [](ai_engine::lm::speech::RecognitionResult result){
        });
        speechProcessor->speechEngine_->setSynthesizedCallback(
                    [](ai_engine::lm::speech::ResultType result) {
        });

        speechProcessor->speechEngine_->recognizeOnce(params, inputData);
    });

    aisdk_speech_processor_complete_recognizing_complete_audio(
        delegate, invocation, "RecognizingCompleteAudio",
        strlen("RecognizingCompleteAudio"), 0);
    return true;
}

bool onHandleStartSynthesizingAudio(AisdkSpeechProcessor *delegate,
                                    GDBusMethodInvocation *invocation,
                                    gchar *voiceProperties,
                                    gpointer userData) {
    auto *speechProcessor = static_cast<SpeechProcessor *>(userData);

    speechProcessor->speechEngine_->startContinuousSynthesis(voiceProperties);

    aisdk_speech_processor_complete_start_synthesizing_audio(
        delegate, invocation);
    return true;
}

bool onHandleInputContinuousText(AisdkSpeechProcessor *delegate,
                                 GDBusMethodInvocation *invocation,
                                 gchar *text,
                                 gpointer userData) {
    std::shared_ptr<SpeechProcessor> speechProcessor =
            static_cast<SpeechProcessor *>(userData)->getSharedObject();
    std::string textString { text };

    cpr::async([speechProcessor, textString]() {
        speechProcessor->speechEngine_->setSynthesizingCallback(
                    [speechProcessor](ai_engine::lm::speech::SynthesisResult result){
            std::cout << "service---onHandleInputContinuousText--callback" << std::endl;
            {
                std::lock_guard<std::mutex> locker(speechProcessor->mutex_);
                if (speechProcessor->stopped) {
                    return;
                }
            }
            speechProcessor->emitSynthesizingResult(
                speechProcessor, result, "SynthesisAudioResult");
        });
        speechProcessor->speechEngine_->setRecognizedCallback(
                    [](ai_engine::lm::speech::RecognitionResult result){
        });
        speechProcessor->speechEngine_->setSynthesizedCallback(
                    [](ai_engine::lm::speech::ResultType result) {
        });

        speechProcessor->speechEngine_->writeContinuousSynthesisText(textString);
    });

    aisdk_speech_processor_complete_input_continuous_text(
        delegate, invocation, "InputContinuousText", 1, 2,
        "InputContinuousText", 10, 1);
    return true;
}

bool onHandleStopSynthesizingAudio(AisdkSpeechProcessor *delegate,
                                   GDBusMethodInvocation *invocation,
                                   gpointer userData) {
    auto *speechProcessor = static_cast<SpeechProcessor *>(userData);

    speechProcessor->speechEngine_->setSynthesizedCallback(
                [](ai_engine::lm::speech::ResultType result) {
       std::cout << "recognizing text finished" ;
    });

    speechProcessor->speechEngine_->stopContinuousSynthesis();

    aisdk_speech_processor_complete_stop_synthesizing_audio(
        delegate, invocation);
    return true;
}

bool onHandleRecognizingCompleteText(AisdkSpeechProcessor *delegate,
                                     GDBusMethodInvocation *invocation,
                                     gchar *voiceProperties,
                                     gchar *text,
                                     gpointer userData) {
    std::shared_ptr<SpeechProcessor> speechProcessor =
            static_cast<SpeechProcessor *>(userData)->getSharedObject();
    std::string textString { text };

    cpr::async([speechProcessor, voiceProperties, textString]() {
        speechProcessor->speechEngine_->setSynthesizingCallback(
                [speechProcessor](ai_engine::lm::speech::SynthesisResult result){
            std::cout << "service---onHandleRecognizingCompleteText--callback" << std::endl;
            {
                std::lock_guard<std::mutex> locker(speechProcessor->mutex_);
                if (speechProcessor->stopped) {
                    return;
                }
            }
            speechProcessor->emitSynthesizingResult(
                speechProcessor, result, "SynthesisAudioOnceResult");
        });
        speechProcessor->speechEngine_->setSynthesizedCallback(
                    [](ai_engine::lm::speech::ResultType result) {
        });
        speechProcessor->speechEngine_->setRecognizedCallback(
                    [](ai_engine::lm::speech::RecognitionResult result){
        });

        speechProcessor->speechEngine_->synthesizeOnce(voiceProperties, textString);
    });

    aisdk_speech_processor_complete_recognizing_complete_text(
        delegate, invocation, "RecognizingCompleteText", 1, 2,
        "RecognizingCompleteText", 10, 1);
    return true;
}

void SpeechProcessor::connectSignal() {
    g_signal_connect(delegate_, "handle-init",
                     G_CALLBACK(onHandleInitEngine),
                     this);

    g_signal_connect(delegate_, "handle-start-recognizing-audio",
                     G_CALLBACK(onHandleStartRecognizingAudio),
                     this);
    g_signal_connect(delegate_, "handle-input-audio-data",
                     G_CALLBACK(onHandleInputAudioData),
                     this);
    g_signal_connect(delegate_, "handle-stop-recognizing-audio",
                     G_CALLBACK(onHandleStopRecognizingAudio),
                     this);
    g_signal_connect(delegate_, "handle-recognizing-complete-audio",
                     G_CALLBACK(onHandleRecognizingCompleteAudio),
                     this);

    g_signal_connect(delegate_, "handle-start-synthesizing-audio",
                     G_CALLBACK(onHandleStartSynthesizingAudio),
                     this);
    g_signal_connect(delegate_, "handle-input-continuous-text",
                     G_CALLBACK(onHandleInputContinuousText),
                     this);
    g_signal_connect(delegate_, "handle-stop-synthesizing-audio",
                     G_CALLBACK(onHandleStopSynthesizingAudio),
                     this);
    g_signal_connect(delegate_, "handle-recognizing-complete-text",
                     G_CALLBACK(onHandleRecognizingCompleteText),
                     this);
}

void SpeechProcessor::unexportSkeleton() {
    if (delegate_ == nullptr) {
        return;
    }

    if (isExported_) {
        g_dbus_interface_skeleton_unexport_from_connection(
            G_DBUS_INTERFACE_SKELETON(delegate_), &connection_);
    }

    g_object_unref(delegate_);
}




