/*
 * Copyright 2024 KylinSoft Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include "services/textprocessor.h"
#include <iostream>
#include <kylin-ai/ai-engine/large-model/cloudnlpengine.h>
#include "services/threadpool/async.h"
#include "textprocessor.h"
#include "config/engineconfiguration.h"

const std::string TextProcessor::objectPath_ =
    "/org/openkylin/aisdk/textprocessor";

TextProcessor::TextProcessor(GDBusConnection &connection,
    ai_engine::AiEnginePluginManager &aiEnginePluginManager)
    : connection_(connection)
    , aiEnginePluginManager_(aiEnginePluginManager) {
    exportSkeleton();
}

TextProcessor::~TextProcessor() {
    unexportSkeleton();
}

void TextProcessor::stopProcess()
{
    if (nlpEngine_) {
        nlpEngine_->stopChat();
    }

    std::lock_guard<std::mutex> locker(mutex_);
    stopped = true;
}

void TextProcessor::onHandleInitEngine(AisdkTextProcessor *delegate,
    GDBusMethodInvocation *invocation, gchar *engineName, gchar *appId,
    gchar *apiKey, gchar *secretKey, gpointer userData)
{
    std::cout << "On init engine..." <<  engineName << appId << apiKey
              << secretKey << invocation << std::endl;
    TextProcessor *textProcessor = static_cast<TextProcessor *>(userData);
    if (!textProcessor) {
        return;
    }
    textProcessor->promptTemplate_.reset(
        new config::PromptTemplate(config::promptTemplateConfigFile(engineName),
                                   config::promptFilePath(engineName))
    );
    textProcessor->initEngine(engineName, appId, apiKey, secretKey);
}

bool TextProcessor::onHandleChat(AisdkTextProcessor *delegate,
                  GDBusMethodInvocation *invocation,
                  gchar *question, gpointer userData) {

    std::cout << "On handle chat..." << invocation << std::endl;
    std::shared_ptr<TextProcessor> textProcessor = static_cast<TextProcessor *>(userData)->getSharedObject();
    if (!textProcessor) {
        return false;
    }
    aisdk_text_processor_complete_chat(delegate, invocation, "");

    std::string userQuestion = question;
    cpr::async([delegate, userQuestion, textProcessor]() {
        textProcessor->nlpEngine_->setChatResultCallback([delegate, textProcessor](const std::string &result) {
            std::cout << "aisdk_text_processor_complete_chat: \n " << result << std::endl;

            {
                std::lock_guard<std::mutex> locker(textProcessor->mutex_);
                if (textProcessor->stopped) {
                    return;
                }
            }

            g_dbus_connection_emit_signal(&(textProcessor->connection_),
                                          "org.openkylin.aisdk.textprocessor",
                                          "/org/openkylin/aisdk/textprocessor",
                                          "org.openkylin.aisdk.textprocessor",
                                          "ChatResult",
                                          g_variant_new("(s)", result.c_str()), nullptr);
        });


        std::string message;
        {
            std::lock_guard<std::mutex> locker(textProcessor->mutex_);
            message = textProcessor->actorChanged ?
                          textProcessor->currentPrompt_ + ":" + userQuestion : userQuestion;
            textProcessor->actorChanged = false;
        }

        textProcessor->nlpEngine_->chat(message);
    });

    return true;
}

void TextProcessor::onSetActor(AisdkTextProcessor *delegate,
           GDBusMethodInvocation *invocation, gint actor,
                gpointer userData) {
    std::cout << "On Set Actor..." << std::endl;
    TextProcessor *textProcessor = static_cast<TextProcessor *>(userData);
    if (!textProcessor) {
        return;
    }

    if (actor == textProcessor->actorId_) {
        return;
    }

    auto prompt = textProcessor->promptTemplate_->prompt((config::PromptTemplate::PromptId)actor);
    std::cout << "Prompt: " << prompt.prompt << std::endl;
    {
        std::lock_guard<std::mutex> locker(textProcessor->mutex_);
        textProcessor->actorChanged = true;
        textProcessor->currentPrompt_ = prompt.prompt;
    }
}

void TextProcessor::onStopChat(AisdkTextProcessor *delegate,
                GDBusMethodInvocation *invocation,
                gpointer userData) {
    std::cout << "On Stop Chat..." << std::endl;
    TextProcessor *textProcessor = static_cast<TextProcessor *>(userData);
    if (!textProcessor) {
        return;
    }

    textProcessor->nlpEngine_->stopChat();
}

void TextProcessor::onSetContextSize(AisdkTextProcessor *delegate,
                GDBusMethodInvocation *invocation, gint size,
                gpointer userData) {
    std::cout << "On Set Context Size..." << std::endl;
    TextProcessor *textProcessor = static_cast<TextProcessor *>(userData);
    if (!textProcessor) {
        return;
    }

    textProcessor->nlpEngine_->setContextSize(size);
}

void TextProcessor::onClearContext(AisdkTextProcessor *delegate,
                      GDBusMethodInvocation *invocation,
                      gpointer userData) {
    std::cout << "On Clear Context..." << std::endl;

    TextProcessor *textProcessor = static_cast<TextProcessor *>(userData);
    if (!textProcessor) {
        return;
    }

    textProcessor->nlpEngine_->clearContext();
}

void TextProcessor::exportSkeleton() {
    delegate_ = aisdk_text_processor_skeleton_new();

    g_signal_connect(delegate_, "handle-init",
                     G_CALLBACK(onHandleInitEngine), this);

    g_signal_connect(delegate_, "handle-chat",
                     G_CALLBACK(onHandleChat), this);

    g_signal_connect(delegate_, "handle-stop-chat",
                     G_CALLBACK(onStopChat), this);

    g_signal_connect(delegate_, "handle-set-actor",
                     G_CALLBACK(onSetActor), this);

    g_signal_connect(delegate_, "handle-set-context",
                     G_CALLBACK(onSetContextSize), this);

    g_signal_connect(delegate_, "handle-clear-context",
                     G_CALLBACK(onClearContext), this);

    GError *error = nullptr;

    isExported_ = g_dbus_interface_skeleton_export(
        G_DBUS_INTERFACE_SKELETON(delegate_), &connection_, objectPath_.c_str(),
        &error);
    if (!isExported_) {
        g_printerr("Error exporting skeleton at address %s: %s\n",
                   objectPath_.c_str(), error->message);
        g_error_free(error);
    }
}

void TextProcessor::unexportSkeleton() {
    if (delegate_ == nullptr) {
        return;
    }

    if (isExported_) {
        g_dbus_interface_skeleton_unexport_from_connection(
            G_DBUS_INTERFACE_SKELETON(delegate_), &connection_);
    }

    g_object_unref(delegate_);
}

void TextProcessor::initEngine(const std::string &engineName,
    const std::string &appId, const std::string &apiKey, const std::string &secretKey)
{
    nlpEngine_ = aiEnginePluginManager_.createNlpEngine(engineName);
    if (nlpEngine_->isCloud()) {
        static_cast<ai_engine::lm::nlp::AbstractCloudNlpEngine *>(nlpEngine_.get())->setAppId(appId);
        static_cast<ai_engine::lm::nlp::AbstractCloudNlpEngine *>(nlpEngine_.get())->setApiKey(apiKey);
        static_cast<ai_engine::lm::nlp::AbstractCloudNlpEngine *>(nlpEngine_.get())->setSecretKey(secretKey);
    }
}
