/*
 * Copyright 2024 KylinSoft Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include "services/textprocessorproxy.h"

#include "services/serverproxy.h"
#include "services/textprocessor.h"

TextProcessorProxy::TextProcessorProxy() { init(); }

TextProcessorProxy::~TextProcessorProxy() { destroy(); }

void TextProcessorProxy::chat(const std::string &question) {
    aisdk_text_processor_call_chat(delegate_, question.c_str(), nullptr,
                                   reinterpret_cast<GAsyncReadyCallback>(
                                       TextProcessorProxy::onChatResultReady),
                                   this);
}

void TextProcessorProxy::init() {
    const auto &unixPath = ServerProxy::getUnixPath();
    GError *error = nullptr;
    if (!ServerProxy::getInstance().available()) {
        g_printerr("Error creating text processor proxy: Server proxy "
                   "connection is unavailable.\n");
        return;
    }

    auto *connection = ServerProxy::getInstance().getConnection();
    const auto &objectPath = TextProcessor::getObjectPath();
    delegate_ = aisdk_text_processor_proxy_new_sync(
        connection, G_DBUS_PROXY_FLAGS_NONE, nullptr, objectPath.c_str(),
        nullptr, &error);
    if (delegate_ == nullptr) {
        g_printerr("Error creating text processor proxy %s: %s\n",
                   objectPath.c_str(), error->message);
        g_error_free(error);
    }
}

void TextProcessorProxy::destroy() {
    if (delegate_ != nullptr) {
        g_object_unref(delegate_);
    }
}

void TextProcessorProxy::onChatResultReady(AisdkTextProcessor *delegate,
                                           GAsyncResult *result,
                                           gpointer userData) {
    auto *proxy = static_cast<TextProcessorProxy *>(userData);
    if (proxy->chatCallback == nullptr) {
        return;
    }

    gchar *answer = nullptr;
    GError *error = nullptr;
    bool success = aisdk_text_processor_call_chat_finish(delegate, &answer,
                                                         result, &error);
    if (success) {
        proxy->chatCallback(answer, proxy->userData_);
        g_free(answer);
    } else {
        proxy->chatCallback("", proxy->userData_);
        g_printerr("Error get text processor proxy chat result: %s\n",
                   error->message);

        g_error_free(error);
    }
}