#include "services/visionprocessor.h"

#include <iostream>
#include <kylin-ai/ai-engine/large-model/cloudvisionengine.h>
#include "config/engineconfiguration.h"
#include "services/threadpool/async.h"

const std::string VisionProcessor::objectPath_ =
    "/org/openkylin/aisdk/visionprocessor";

VisionProcessor::VisionProcessor(
    GDBusConnection &connection,
    ai_engine::AiEnginePluginManager &aiEnginePluginManager)
    : connection_(connection), aiEnginePluginManager_(aiEnginePluginManager) {
    exportSkeleton();
}

VisionProcessor::~VisionProcessor() { unexportSkeleton(); }

void VisionProcessor::exportSkeleton() {
    delegate_ = aisdk_vision_processor_skeleton_new();
    connectSignal();

    GError *error = nullptr;

    isExported_ = g_dbus_interface_skeleton_export(
        G_DBUS_INTERFACE_SKELETON(delegate_), &connection_, objectPath_.c_str(),
        &error);
    if (!isExported_) {
        g_printerr("Error creating server at address %s: %s\n",
                   objectPath_.c_str(), error->message);
        g_error_free(error);
    }
}

void VisionProcessor::initEngine(const std::string &engineName, const std::string &appId,
                const std::string &apiKey, const std::string &secretKey) {

    visionEngine_ = aiEnginePluginManager_.createVisionEngine(engineName);
    if (visionEngine_->isCloud()) {
        static_cast<ai_engine::lm::vision::AbstractCloudVisionEngine *>(visionEngine_.get())->setAppId(appId);
        static_cast<ai_engine::lm::vision::AbstractCloudVisionEngine *>(visionEngine_.get())->setApiKey(apiKey);
        static_cast<ai_engine::lm::vision::AbstractCloudVisionEngine *>(visionEngine_.get())->setSecretKey(secretKey);
    }
}

void onHandleInitEngine(AisdkVisionProcessor *delegate,
    GDBusMethodInvocation *invocation, gchar *engineName, gchar *appId,
    gchar *apiKey, gchar *secretKey, gpointer userData) {
    std::cout << "service---onHandleInitEngine---" <<  engineName << appId << apiKey
              << secretKey << invocation << std::endl;
    auto *visionProcessor = static_cast<VisionProcessor *>(userData);
    if (!visionProcessor) {
        return;
    }
    visionProcessor->promptTemplate_.reset(
        new config::PromptTemplate(config::promptTemplateConfigFile(engineName),
                                   config::promptFilePath(engineName))
    );
    visionProcessor->initEngine(engineName, appId, apiKey, secretKey);
}

bool onHandleSetPromptImageSize(AisdkVisionProcessor *delegate,
                                GDBusMethodInvocation *invocation,
                                gint imageWidth,
                                gint imageHeight,
                                gpointer userData) {
    std::cout << "service---onHandleSetPromptImageSize---" << std::endl;
    auto *visionProcessor = static_cast<VisionProcessor *>(userData);
    visionProcessor->visionEngine_->setPrompt2ImageSize(imageWidth, imageHeight);

    aisdk_vision_processor_complete_set_prompt_image_size(
        delegate, invocation
    );
    return true;
}

bool onHandleSetPromptImageNumber(AisdkVisionProcessor *delegate,
                                  GDBusMethodInvocation *invocation,
                                  gint imageNumber,
                                  gpointer userData) {
    std::cout << "service---onHandleSetPromptImageNumber---" << std::endl;
    auto *visionProcessor = static_cast<VisionProcessor *>(userData);
    visionProcessor->visionEngine_->setPrompt2ImageNumber(imageNumber);

    aisdk_vision_processor_complete_set_prompt_image_number(
        delegate, invocation);
    return true;
}

bool onHandleSetPromptImageStyle(AisdkVisionProcessor *delegate,
                                 GDBusMethodInvocation *invocation,
                                 gint imageStyle,
                                 gpointer userData) {
    std::cout << "service---onHandleSetPromptImageStyle---" << std::endl;
    auto *visionProcessor = static_cast<VisionProcessor *>(userData);
    visionProcessor->visionEngine_->setPrompt2ImageStyle(imageStyle);

    aisdk_vision_processor_complete_set_prompt_image_style(
        delegate, invocation);
    return true;
}

bool onHandlePromptImage(AisdkVisionProcessor *delegate,
                         GDBusMethodInvocation *invocation,
                         const gchar *prompt,
                         gpointer userData) {
    std::cout << "service---onHandlePromptImage---" << std::endl;
    auto *visionProcessor = static_cast<VisionProcessor *>(userData);

    std::string promptString{prompt};
    cpr::async([visionProcessor, promptString]() {
        visionProcessor->visionEngine_->setPrompt2ImageCallback(
                    [visionProcessor](ai_engine::lm::vision::ImageData imageData){
            std::cout << "service---onHandlePromptImage--callback-" << std::endl;
            std::cout << "service--Imagecallback--data--" << imageData.imageData.size() << std::endl;
            std::cout << "service--Imagecallback--width--" << imageData.width << std::endl;
            {
                std::lock_guard<std::mutex> locker(visionProcessor->mutex_);
                if (visionProcessor->stopped) {
                    return;
                }
            }

            // 创建包含字节数组的 GVariant
            GVariant *byte_array_variant = g_variant_new_from_data(
                G_VARIANT_TYPE_BYTESTRING,
                imageData.imageData.data(),
                imageData.imageData.size(),
                TRUE, NULL, NULL);

            GVariantBuilder builder;
            g_variant_builder_init (&builder, G_VARIANT_TYPE_TUPLE);
            g_variant_builder_open (&builder, G_VARIANT_TYPE_ARRAY);
            g_variant_builder_add_value(&builder, byte_array_variant);
            g_variant_builder_close (&builder);

            g_variant_builder_add(&builder, "i", imageData.format);
            g_variant_builder_add(&builder, "i", imageData.height);
            g_variant_builder_add(&builder, "i", imageData.width);
            g_variant_builder_add(&builder, "i", imageData.total);
            g_variant_builder_add(&builder, "i", imageData.index);
            g_variant_builder_add(&builder, "i", imageData.errorCode);

            g_dbus_connection_emit_signal(&(visionProcessor->connection_),
                                          "org.openkylin.aisdk.visionprocessor",
                                          "/org/openkylin/aisdk/visionprocessor",
                                          "org.openkylin.aisdk.visionprocessor",
                                          "PromptImageResult",
                                          g_variant_builder_end(&builder),
                                          nullptr);
        });

        visionProcessor->visionEngine_->prompt2Image(promptString);
    });

    aisdk_vision_processor_complete_prompt_image(
        delegate, invocation, 1,2,3,4,5,"prompt_image",10, 1);

    return true;
}

void VisionProcessor::connectSignal() {
    g_signal_connect(delegate_, "handle-init",
                     G_CALLBACK(onHandleInitEngine),
                     this);
    g_signal_connect(delegate_, "handle-set-prompt-image-size",
                     G_CALLBACK(onHandleSetPromptImageSize),
                     this);
    g_signal_connect(delegate_, "handle-set-prompt-image-number",
                     G_CALLBACK(onHandleSetPromptImageNumber),
                     this);
    g_signal_connect(delegate_, "handle-set-prompt-image-style",
                     G_CALLBACK(onHandleSetPromptImageStyle),
                     this);
    g_signal_connect(delegate_, "handle-prompt-image",
                     G_CALLBACK(onHandlePromptImage),
                     this);                     
}

void VisionProcessor::unexportSkeleton() {
    if (delegate_ == nullptr) {
        return;
    }

    if (isExported_) {
        g_dbus_interface_skeleton_unexport_from_connection(
            G_DBUS_INTERFACE_SKELETON(delegate_), &connection_);
    }

    g_object_unref(delegate_);
}



